#include "esp_camera.h"
#include <WiFi.h>
#include <driver/adc.h>
#include "FS.h"
#include "SD.h"
#include "SPI.h"
#include <Wire.h>
#include <DFRobot_DHT20.h>

//
// WARNING!!! PSRAM IC required for UXGA resolution and high JPEG quality
//            Ensure ESP32 Wrover Module or other board with PSRAM is selected
//            Partial images will be transmitted if image exceeds buffer size
//

// Select camera model
#define CAMERA_MODEL_AI_THINKER // Has PSRAM

#define PWDN_GPIO_NUM    -1
#define RESET_GPIO_NUM   12
#define XCLK_GPIO_NUM     4
#define SIOD_GPIO_NUM    17
#define SIOC_GPIO_NUM    18

#define Y9_GPIO_NUM      11
#define Y8_GPIO_NUM      10
#define Y7_GPIO_NUM       9
#define Y6_GPIO_NUM      46
#define Y5_GPIO_NUM       3
#define Y4_GPIO_NUM       8
#define Y3_GPIO_NUM      16
#define Y2_GPIO_NUM      15
#define VSYNC_GPIO_NUM    5
#define HREF_GPIO_NUM     6
#define PCLK_GPIO_NUM     7

#define	FREQ1			 1
#define	AN13			 13

#define LED_GREEN	42
#define LED_RED	 	41

#define LEDG_ON		digitalWrite( LED_GREEN, HIGH)
#define LEDG_OFF	digitalWrite( LED_GREEN, LOW)
#define LEDR_ON		digitalWrite( LED_RED, HIGH)
#define LEDR_OFF	digitalWrite( LED_RED, LOW)

const char* ssid = "AirPort-----";
const char* password = "1620353------";
const IPAddress 	subnet(255,255,255,0);
const IPAddress 	fix_ip(192,168,0,13);

#define	VSPI_SCK	21
#define	VSPI_MOSI	47
#define	VSPI_MISO	14
#define	VSPI_SS		48
#define	SPI_FREQ	40000000

#define	I2C_SDA			1
#define	I2C_SCL			2	

float	temperature;
float	humidity;

DFRobot_DHT20 dht20;

void startCameraServer();
int	 index_buf_len;

#if CONFIG_IDF_TARGET_ESP32S2 || CONFIG_IDF_TARGET_ESP32S3
#define VSPI FSPI
#endif

SPIClass * vspi = NULL;

void	set_freq( int f)
{
	ledcSetup( 0, 50, 10);		// 0ch 100Hz 12bit
	ledcAttachPin( FREQ1, 0);	// pin 0ch
  	ledcWrite( 0, 1024-f);		// 0ch pulse width
}

void	led_init( void)
{
	pinMode(LED_GREEN, OUTPUT);
	pinMode(LED_RED, OUTPUT);
	LEDR_ON;
	LEDG_OFF;
}

void listDir(fs::FS &fs, const char * dirname, uint8_t levels){
    Serial.printf("Listing directory: %s\n", dirname);

    File root = fs.open(dirname);
    if(!root){
        Serial.println("Failed to open directory");
        return;
    }
    if(!root.isDirectory()){
        Serial.println("Not a directory");
        return;
    }

    File file = root.openNextFile();
    while(file){
        if(file.isDirectory()){
            Serial.print("  DIR : ");
            Serial.println(file.name());
            if(levels){
                listDir(fs, file.path(), levels -1);
            }
        } else {
            Serial.print("  FILE: ");
            Serial.print(file.name());
            Serial.print("  SIZE: ");
            Serial.println(file.size());
        }
        file = root.openNextFile();
    }
}

void createDir(fs::FS &fs, const char * path){
    Serial.printf("Creating Dir: %s\n", path);
    if(fs.mkdir(path)){
        Serial.println("Dir created");
    } else {
        Serial.println("mkdir failed");
    }
}

void removeDir(fs::FS &fs, const char * path){
    Serial.printf("Removing Dir: %s\n", path);
    if(fs.rmdir(path)){
        Serial.println("Dir removed");
    } else {
        Serial.println("rmdir failed");
    }
}

void readFile(fs::FS &fs, const char * path){
    Serial.printf("Reading file: %s\n", path);

    File file = fs.open(path);
    if(!file){
        Serial.println("Failed to open file for reading");
        return;
    }

    Serial.print("Read from file: ");
    while(file.available()){
        Serial.write(file.read());
    }
    file.close();
}

void writeFile(fs::FS &fs, const char * path, const char * message){
    Serial.printf("Writing file: %s\n", path);

    File file = fs.open(path, FILE_WRITE);
    if(!file){
        Serial.println("Failed to open file for writing");
        return;
    }
    if(file.print(message)){
        Serial.println("File written");
    } else {
        Serial.println("Write failed");
    }
    file.close();
}

void appendFile(fs::FS &fs, const char * path, const char * message){
    Serial.printf("Appending to file: %s\n", path);

    File file = fs.open(path, FILE_APPEND);
    if(!file){
        Serial.println("Failed to open file for appending");
        return;
    }
    if(file.print(message)){
        Serial.println("Message appended");
    } else {
        Serial.println("Append failed");
    }
    file.close();
}

void renameFile(fs::FS &fs, const char * path1, const char * path2){
    Serial.printf("Renaming file %s to %s\n", path1, path2);
    if (fs.rename(path1, path2)) {
        Serial.println("File renamed");
    } else {
        Serial.println("Rename failed");
    }
}

void deleteFile(fs::FS &fs, const char * path){
    Serial.printf("Deleting file: %s\n", path);
    if(fs.remove(path)){
        Serial.println("File deleted");
    } else {
        Serial.println("Delete failed");
    }
}

void testFileIO(fs::FS &fs, const char * path){
    File file = fs.open(path);
    static uint8_t buf[512];
    size_t len = 0;
    uint32_t start = millis();
    uint32_t end = start;
    if(file){
        len = file.size();
        size_t flen = len;
        start = millis();
        while(len){
            size_t toRead = len;
            if(toRead > 512){
                toRead = 512;
            }
            file.read(buf, toRead);
            len -= toRead;
        }
        end = millis() - start;
        Serial.printf("%u bytes read for %u ms\n", flen, end);
        file.close();
    } else {
        Serial.println("Failed to open file for reading");
    }


    file = fs.open(path, FILE_WRITE);
    if(!file){
        Serial.println("Failed to open file for writing");
        return;
    }

    size_t i;
    start = millis();
    for(i=0; i<2048; i++){
        file.write(buf, 512);
    }
    end = millis() - start;
    Serial.printf("%u bytes written for %u ms\n", 2048 * 512, end);
    file.close();
}

int sdc_init( void)
{
	int	spi_ok;
	
#ifdef	CONFIG_IDF_TARGET_ESP32S3
 	vspi = new SPIClass( VSPI);
 	if( vspi != NULL){
		vspi->begin( VSPI_SCK, VSPI_MISO, VSPI_MOSI, VSPI_SS);
		spi_ok = 1;
 	} else {
 		spi_ok = 0;
 	}
#else
	return 0;
#endif
	if( spi_ok){
		Serial.println(" VSPI Start");
	   if(!SD.begin( VSPI_SS, *vspi, SPI_FREQ)){
		   if(!SD.begin( VSPI_SS, *vspi, SPI_FREQ)){	// need 1 retry after power on
	     	   Serial.println("Card Mount Failed");
	        	return 0;
		   }
	    }
	    uint8_t cardType = SD.cardType();
	
	    if(cardType == CARD_NONE){
	        Serial.println("No SD card attached");
	        return 0;
	    }
	
	    Serial.print("SD Card Type: ");
	    if(cardType == CARD_MMC){
	        Serial.println("MMC");
	    } else if(cardType == CARD_SD){
	        Serial.println("SDSC");
	    } else if(cardType == CARD_SDHC){
	        Serial.println("SDHC");
	    } else {
	        Serial.println("UNKNOWN");
	    }
	
	    uint64_t cardSize = SD.cardSize() / (1024 * 1024);
	    Serial.printf("SD Card Size: %lluMB\n", cardSize);
	
	    listDir(SD, "/", 0);
	    createDir(SD, "/mydir");
	    listDir(SD, "/", 0);
	    removeDir(SD, "/mydir");
	    listDir(SD, "/", 2);
	    writeFile(SD, "/hello.txt", "Hello ");
	    appendFile(SD, "/hello.txt", "World!\n");
	    readFile(SD, "/hello.txt");
	    deleteFile(SD, "/foo.txt");
	    renameFile(SD, "/hello.txt", "/foo.txt");
	    readFile(SD, "/foo.txt");
	    testFileIO(SD, "/test.txt");
	    Serial.printf("Total space: %lluMB\n", SD.totalBytes() / (1024 * 1024));
	    Serial.printf("Used space: %lluMB\n", SD.usedBytes() / (1024 * 1024));
	    LEDG_ON;
	    return 1;
	} else {
		return 0;
	}
}

void	i2c_init( void)
{
	Wire.begin( I2C_SDA, I2C_SCL);
}

void	init_DHT20( void)
{
	while( dht20.begin()){
	    Serial.println("Initialize sensor failed");
	    delay(1000);
	}
}


void setup( void)
{

	Serial.begin(115200);
	Serial.setDebugOutput(true);
	Serial.println();

	led_init( );

	camera_config_t config;
	config.pin_d0 = Y2_GPIO_NUM;
	config.pin_d1 = Y3_GPIO_NUM;
	config.pin_d2 = Y4_GPIO_NUM;
	config.pin_d3 = Y5_GPIO_NUM;
	config.pin_d4 = Y6_GPIO_NUM;
	config.pin_d5 = Y7_GPIO_NUM;
	config.pin_d6 = Y8_GPIO_NUM;
	config.pin_d7 = Y9_GPIO_NUM;
	config.pin_xclk = XCLK_GPIO_NUM;
	config.pin_pclk = PCLK_GPIO_NUM;
	config.pin_vsync = VSYNC_GPIO_NUM;
	config.pin_href = HREF_GPIO_NUM;
	config.pin_sscb_sda = SIOD_GPIO_NUM;
	config.pin_sscb_scl = SIOC_GPIO_NUM;
	config.pin_pwdn = PWDN_GPIO_NUM;
	config.pin_reset = RESET_GPIO_NUM;
	config.xclk_freq_hz = 14000000;
	config.pixel_format = PIXFORMAT_JPEG;

	// if PSRAM IC present, init with UXGA resolution and higher JPEG quality
	//											for larger pre-allocated frame buffer.
	if( psramFound()){
		config.frame_size = FRAMESIZE_SVGA;	// FRAMESIZE_UXGA;
		config.jpeg_quality = 12;			// 10;
		config.fb_count = 1;				// 2;
	} else {
		config.frame_size = FRAMESIZE_SVGA;
		config.jpeg_quality = 12;
		config.fb_count = 1;
	}

	// camera init
	esp_err_t err = esp_camera_init(&config);
	if (err != ESP_OK) {
		Serial.printf("Camera init failed with error 0x%x", err);
		return;
	}

	sensor_t * s = esp_camera_sensor_get();
	// initial sensors are flipped vertically and colors are a bit saturated
	if( s->id.PID == OV2640_PID){
		Serial.printf("Camera OV2640 detected");
	}

#if defined(CAMERA_MODEL_M5STACK_WIDE) || defined(CAMERA_MODEL_M5STACK_ESP32CAM)
	s->set_vflip(s, 1);
	s->set_hmirror(s, 1);
#endif

	WiFi.begin(ssid, password);
	WiFi.config( fix_ip, fix_ip,subnet);

	while (WiFi.status() != WL_CONNECTED) {
		delay(500);
		Serial.print(".");
	}
	Serial.println("");
	Serial.println("WiFi connected");
	LEDR_OFF;

	startCameraServer();

	Serial.print("Camera Ready! Use 'http://");
	Serial.print(WiFi.localIP());
	Serial.println("' to connect");

	i2c_init();
	sdc_init();

	init_DHT20();
}

void loop() {
 	delay(100);
 	temperature = dht20.getTemperature();
 	humidity	= dht20.getHumidity();
 }
