// Copyright 2015-2016 Espressif Systems (Shanghai) PTE LTD
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
#include "esp_http_server.h"
#include "esp_timer.h"
#include "esp_camera.h"
#include "img_converters.h"
#include "fb_gfx.h"
#include "driver/ledc.h"
#include "sdkconfig.h"

#if defined(ARDUINO_ARCH_ESP32) && defined(CONFIG_ARDUHAL_ESP_LOG)
#include "esp32-hal-log.h"
#define TAG ""
#else
#include "esp_log.h"
static const char *TAG = "camera_httpd";
#endif

extern	float	temperature;
extern	float	humidity;
typedef struct
{
    httpd_req_t *req;
    size_t len;
} jpg_chunking_t;

#define PART_BOUNDARY "123456789000000000000987654321"
static const char *_STREAM_CONTENT_TYPE = "multipart/x-mixed-replace;boundary=" PART_BOUNDARY;
static const char *_STREAM_BOUNDARY = "\r\n--" PART_BOUNDARY "\r\n";
static const char *_STREAM_PART = "Content-Type: image/jpeg\r\nContent-Length: %u\r\nX-Timestamp: %d.%06d\r\n\r\n";

httpd_handle_t stream_httpd = NULL;
httpd_handle_t camera_httpd = NULL;


typedef struct
{
    size_t size;  //number of values used for filtering
    size_t index; //current value index
    size_t count; //value count
    int sum;
    int *values; //array to be filled with values
} ra_filter_t;

static ra_filter_t ra_filter;

static ra_filter_t *ra_filter_init(ra_filter_t *filter, size_t sample_size)
{
    memset(filter, 0, sizeof(ra_filter_t));

    filter->values = (int *)malloc(sample_size * sizeof(int));
    if (!filter->values)
    {
        return NULL;
    }
    memset(filter->values, 0, sample_size * sizeof(int));

    filter->size = sample_size;
    return filter;
}

static int ra_filter_run(ra_filter_t *filter, int value)
{
    if (!filter->values)
    {
        return value;
    }
    filter->sum -= filter->values[filter->index];
    filter->values[filter->index] = value;
    filter->sum += filter->values[filter->index];
    filter->index++;
    filter->index = filter->index % filter->size;
    if (filter->count < filter->size)
    {
        filter->count++;
    }
    return filter->sum / filter->count;
}

static esp_err_t bmp_handler(httpd_req_t *req)
{
    camera_fb_t *fb = NULL;
    esp_err_t res = ESP_OK;
    uint64_t fr_start = esp_timer_get_time();
    fb = esp_camera_fb_get();
    if (!fb)
    {
        ESP_LOGE(TAG, "Camera capture failed");
        httpd_resp_send_500(req);
        return ESP_FAIL;
    }

    httpd_resp_set_type(req, "image/x-windows-bmp");
    httpd_resp_set_hdr(req, "Content-Disposition", "inline; filename=capture.bmp");
    httpd_resp_set_hdr(req, "Access-Control-Allow-Origin", "*");

    char ts[32];
    snprintf(ts, 32, "%ld.%06ld", fb->timestamp.tv_sec, fb->timestamp.tv_usec);
    httpd_resp_set_hdr(req, "X-Timestamp", (const char *)ts);


    uint8_t * buf = NULL;
    size_t buf_len = 0;
    bool converted = frame2bmp(fb, &buf, &buf_len);
    esp_camera_fb_return(fb);
    if(!converted){
        ESP_LOGE(TAG, "BMP Conversion failed");
        httpd_resp_send_500(req);
        return ESP_FAIL;
    }
    res = httpd_resp_send(req, (const char *)buf, buf_len);
    free(buf);
    uint64_t fr_end = esp_timer_get_time();
    ESP_LOGI(TAG, "BMP: %llums, %uB", (uint64_t)((fr_end - fr_start) / 1000), buf_len);
    return res;
}

static size_t jpg_encode_stream(void *arg, size_t index, const void *data, size_t len)
{
    jpg_chunking_t *j = (jpg_chunking_t *)arg;
    if (!index)
    {
        j->len = 0;
    }
    if (httpd_resp_send_chunk(j->req, (const char *)data, len) != ESP_OK)
    {
        return 0;
    }
    j->len += len;
    return len;
}

static esp_err_t capture_handler(httpd_req_t *req)
{
    camera_fb_t *fb = NULL;
    esp_err_t res = ESP_OK;
    int64_t fr_start = esp_timer_get_time();

    fb = esp_camera_fb_get();

    if (!fb)
    {
        ESP_LOGE(TAG, "Camera capture failed");
        httpd_resp_send_500(req);
        return ESP_FAIL;
    }

    httpd_resp_set_type(req, "image/jpeg");
    httpd_resp_set_hdr(req, "Content-Disposition", "inline; filename=capture.jpg");
    httpd_resp_set_hdr(req, "Access-Control-Allow-Origin", "*");

    char ts[32];
    snprintf(ts, 32, "%ld.%06ld", fb->timestamp.tv_sec, fb->timestamp.tv_usec);
    httpd_resp_set_hdr(req, "X-Timestamp", (const char *)ts);

        size_t fb_len = 0;
        if (fb->format == PIXFORMAT_JPEG)
        {
            fb_len = fb->len;
            res = httpd_resp_send(req, (const char *)fb->buf, fb->len);
        }
        else
        {
            jpg_chunking_t jchunk = {req, 0};
            res = frame2jpg_cb(fb, 80, jpg_encode_stream, &jchunk) ? ESP_OK : ESP_FAIL;
            httpd_resp_send_chunk(req, NULL, 0);
            fb_len = jchunk.len;
        }
        esp_camera_fb_return(fb);
        int64_t fr_end = esp_timer_get_time();
        ESP_LOGI(TAG, "JPG: %uB %ums", (uint32_t)(fb_len), (uint32_t)((fr_end - fr_start) / 1000));
        return res;
}

static esp_err_t stream_handler(httpd_req_t *req)
{
    camera_fb_t *fb = NULL;
    struct timeval _timestamp;
    esp_err_t res = ESP_OK;
    size_t _jpg_buf_len = 0;
    uint8_t *_jpg_buf = NULL;
    char *part_buf[128];

    static int64_t last_frame = 0;
    if (!last_frame)
    {
        last_frame = esp_timer_get_time();
    }

    res = httpd_resp_set_type(req, _STREAM_CONTENT_TYPE);
    if (res != ESP_OK)
    {
        return res;
    }

    httpd_resp_set_hdr(req, "Access-Control-Allow-Origin", "*");
    httpd_resp_set_hdr(req, "X-Framerate", "60");


    while (true)
    {

        fb = esp_camera_fb_get();
        if (!fb)
        {
            ESP_LOGE(TAG, "Camera capture failed");
            res = ESP_FAIL;
        }
        else
        {
            _timestamp.tv_sec = fb->timestamp.tv_sec;
            _timestamp.tv_usec = fb->timestamp.tv_usec;
                if (fb->format != PIXFORMAT_JPEG)
                {
                    bool jpeg_converted = frame2jpg(fb, 80, &_jpg_buf, &_jpg_buf_len);
                    esp_camera_fb_return(fb);
                    fb = NULL;
                    if (!jpeg_converted)
                    {
                        ESP_LOGE(TAG, "JPEG compression failed");
                        res = ESP_FAIL;
                    }
                }
                else
                {
                    _jpg_buf_len = fb->len;
                    _jpg_buf = fb->buf;
                }
        }
        if (res == ESP_OK)
        {
            res = httpd_resp_send_chunk(req, _STREAM_BOUNDARY, strlen(_STREAM_BOUNDARY));
        }
        if (res == ESP_OK)
        {
            size_t hlen = snprintf((char *)part_buf, 128, _STREAM_PART, _jpg_buf_len, _timestamp.tv_sec, _timestamp.tv_usec);
            res = httpd_resp_send_chunk(req, (const char *)part_buf, hlen);
        }
        if (res == ESP_OK)
        {
            res = httpd_resp_send_chunk(req, (const char *)_jpg_buf, _jpg_buf_len);
        }
        if (fb)
        {
            esp_camera_fb_return(fb);
            fb = NULL;
            _jpg_buf = NULL;
        }
        else if (_jpg_buf)
        {
            free(_jpg_buf);
            _jpg_buf = NULL;
        }
        if (res != ESP_OK)
        {
            break;
        }
        int64_t fr_end = esp_timer_get_time();


        int64_t frame_time = fr_end - last_frame;
        last_frame = fr_end;
        frame_time /= 1000;
        uint32_t avg_frame_time = ra_filter_run(&ra_filter, frame_time);
        ESP_LOGI(TAG, "MJPG: %uB %ums (%.1ffps), AVG: %ums (%.1ffps)"
                 ,
                 (uint32_t)(_jpg_buf_len),
                 (uint32_t)frame_time, 1000.0 / (uint32_t)frame_time,
                 avg_frame_time, 1000.0 / avg_frame_time
        );
    }
    last_frame = 0;
    return res;
}

static esp_err_t parse_get(httpd_req_t *req, char **obuf)
{
    char *buf = NULL;
    size_t buf_len = 0;

    buf_len = httpd_req_get_url_query_len(req) + 1;
    if (buf_len > 1) {
        buf = (char *)malloc(buf_len);
        if (!buf) {
            httpd_resp_send_500(req);
            return ESP_FAIL;
        }
        if (httpd_req_get_url_query_str(req, buf, buf_len) == ESP_OK) {
            *obuf = buf;
            return ESP_OK;
        }
        free(buf);
    }
    httpd_resp_send_404(req);
    return ESP_FAIL;
}

static int print_reg(char * p, sensor_t * s, uint16_t reg, uint32_t mask){
    return sprintf(p, "\"0x%x\":%u,", reg, s->get_reg(s, reg, mask));
}

static esp_err_t status_handler(httpd_req_t *req)
{
    static char json_response[1024];

    sensor_t *s = esp_camera_sensor_get();
    char *p = json_response;
    *p++ = '{';

    if(s->id.PID == OV5640_PID || s->id.PID == OV3660_PID){
        for(int reg = 0x3400; reg < 0x3406; reg+=2){
            p+=print_reg(p, s, reg, 0xFFF);//12 bit
        }
        p+=print_reg(p, s, 0x3406, 0xFF);

        p+=print_reg(p, s, 0x3500, 0xFFFF0);//16 bit
        p+=print_reg(p, s, 0x3503, 0xFF);
        p+=print_reg(p, s, 0x350a, 0x3FF);//10 bit
        p+=print_reg(p, s, 0x350c, 0xFFFF);//16 bit

        for(int reg = 0x5480; reg <= 0x5490; reg++){
            p+=print_reg(p, s, reg, 0xFF);
        }

        for(int reg = 0x5380; reg <= 0x538b; reg++){
            p+=print_reg(p, s, reg, 0xFF);
        }

        for(int reg = 0x5580; reg < 0x558a; reg++){
            p+=print_reg(p, s, reg, 0xFF);
        }
        p+=print_reg(p, s, 0x558a, 0x1FF);//9 bit
    } else if(s->id.PID == OV2640_PID){
        p+=print_reg(p, s, 0xd3, 0xFF);
        p+=print_reg(p, s, 0x111, 0xFF);
        p+=print_reg(p, s, 0x132, 0xFF);
    }

    p += sprintf(p, "\"xclk\":%u,", s->xclk_freq_hz / 1000000);
    p += sprintf(p, "\"pixformat\":%u,", s->pixformat);
    p += sprintf(p, "\"framesize\":%u,", s->status.framesize);
    p += sprintf(p, "\"quality\":%u,", s->status.quality);
    p += sprintf(p, "\"brightness\":%d,", s->status.brightness);
    p += sprintf(p, "\"contrast\":%d,", s->status.contrast);
    p += sprintf(p, "\"saturation\":%d,", s->status.saturation);
    p += sprintf(p, "\"sharpness\":%d,", s->status.sharpness);
    p += sprintf(p, "\"special_effect\":%u,", s->status.special_effect);
    p += sprintf(p, "\"wb_mode\":%u,", s->status.wb_mode);
    p += sprintf(p, "\"awb\":%u,", s->status.awb);
    p += sprintf(p, "\"awb_gain\":%u,", s->status.awb_gain);
    p += sprintf(p, "\"aec\":%u,", s->status.aec);
    p += sprintf(p, "\"aec2\":%u,", s->status.aec2);
    p += sprintf(p, "\"ae_level\":%d,", s->status.ae_level);
    p += sprintf(p, "\"aec_value\":%u,", s->status.aec_value);
    p += sprintf(p, "\"agc\":%u,", s->status.agc);
    p += sprintf(p, "\"agc_gain\":%u,", s->status.agc_gain);
    p += sprintf(p, "\"gainceiling\":%u,", s->status.gainceiling);
    p += sprintf(p, "\"bpc\":%u,", s->status.bpc);
    p += sprintf(p, "\"wpc\":%u,", s->status.wpc);
    p += sprintf(p, "\"raw_gma\":%u,", s->status.raw_gma);
    p += sprintf(p, "\"lenc\":%u,", s->status.lenc);
    p += sprintf(p, "\"hmirror\":%u,", s->status.hmirror);
    p += sprintf(p, "\"dcw\":%u,", s->status.dcw);
    p += sprintf(p, "\"colorbar\":%u", s->status.colorbar);
    p += sprintf(p, ",\"led_intensity\":%d", -1);
    *p++ = '}';
    *p++ = 0;
    httpd_resp_set_type(req, "application/json");
    httpd_resp_set_hdr(req, "Access-Control-Allow-Origin", "*");
    return httpd_resp_send(req, json_response, strlen(json_response));
}

char	index_buf[ 65536];


int	get_buf_size( void)
{
	int	i;
		
	for( i = 0 ; i < sizeof( index_buf); i++){
		if( index_buf[ i] == 0){
			break;
		}
	}
	return i;
}

int	get_html( void)
{
	char	b[ 1024];
	char	*p;

	p = index_buf;
	p[ 0] = 0;

	sprintf( b,"<!doctype html>\n");	p = strcat( p, b);
	sprintf( b,"<html>\n");	p = strcat( p, b);
	sprintf( b,"    <head>\n");	p = strcat( p, b);
	sprintf( b,"        <meta charset=\"utf-8\">\n");	p = strcat( p, b);
	sprintf( b,"        <meta name=\"viewport\" content=\"width=device-width,initial-scale=1\">\n");	p = strcat( p, b);
	sprintf( b,"        <title>ESP32 OV2460 Camera Test</title>\n");	p = strcat( p, b);
	sprintf( b,"        <style>\n");	p = strcat( p, b);
	sprintf( b,"            body {\n");	p = strcat( p, b);
	sprintf( b,"                font-family: Arial,Helvetica,sans-serif;\n");	p = strcat( p, b);
	sprintf( b,"                background: #066361;\n");	p = strcat( p, b);
	sprintf( b,"                color: #EFEFEF;\n");	p = strcat( p, b);
	sprintf( b,"                font-size: 16px\n");	p = strcat( p, b);
	sprintf( b,"            }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"            h2 {\n");	p = strcat( p, b);
	sprintf( b,"                font-size: 18px\n");	p = strcat( p, b);
	sprintf( b,"            }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"            section.main {\n");	p = strcat( p, b);
	sprintf( b,"                display: flex\n");	p = strcat( p, b);
	sprintf( b,"            }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"            #menu,section.main {\n");	p = strcat( p, b);
	sprintf( b,"                flex-direction: column\n");	p = strcat( p, b);
	sprintf( b,"            }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"            #menu {\n");	p = strcat( p, b);
	sprintf( b,"                display: none;\n");	p = strcat( p, b);
	sprintf( b,"                flex-wrap: nowrap;\n");	p = strcat( p, b);
	sprintf( b,"                min-width: 220px;\n");	p = strcat( p, b);
	sprintf( b,"                background: #363636;\n");	p = strcat( p, b);
	sprintf( b,"                padding: 8px;\n");	p = strcat( p, b);
	sprintf( b,"                border-radius: 4px;\n");	p = strcat( p, b);
	sprintf( b,"                margin-top: -10px;\n");	p = strcat( p, b);
	sprintf( b,"                margin-right: 10px;\n");	p = strcat( p, b);
	sprintf( b,"            }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"            #content {\n");	p = strcat( p, b);
	sprintf( b,"                display: flex;\n");	p = strcat( p, b);
	sprintf( b,"                flex-wrap: wrap;\n");	p = strcat( p, b);
	sprintf( b,"                align-items: stretch\n");	p = strcat( p, b);
	sprintf( b,"            }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"            figure {\n");	p = strcat( p, b);
	sprintf( b,"                padding: 0px;\n");	p = strcat( p, b);
	sprintf( b,"                margin: 0;\n");	p = strcat( p, b);
	sprintf( b,"                -webkit-margin-before: 0;\n");	p = strcat( p, b);
	sprintf( b,"                margin-block-start: 0;\n");	p = strcat( p, b);
	sprintf( b,"                -webkit-margin-after: 0;\n");	p = strcat( p, b);
	sprintf( b,"                margin-block-end: 0;\n");	p = strcat( p, b);
	sprintf( b,"                -webkit-margin-start: 0;\n");	p = strcat( p, b);
	sprintf( b,"                margin-inline-start: 0;\n");	p = strcat( p, b);
	sprintf( b,"                -webkit-margin-end: 0;\n");	p = strcat( p, b);
	sprintf( b,"                margin-inline-end: 0\n");	p = strcat( p, b);
	sprintf( b,"            }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"            figure img {\n");	p = strcat( p, b);
	sprintf( b,"                display: block;\n");	p = strcat( p, b);
	sprintf( b,"                width: 100%;\n");	p = strcat( p, b);
	sprintf( b,"                height: auto;\n");	p = strcat( p, b);
	sprintf( b,"                border-radius: 4px;\n");	p = strcat( p, b);
	sprintf( b,"                margin-top: 8px;\n");	p = strcat( p, b);
	sprintf( b,"            }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"            section#buttons {\n");	p = strcat( p, b);
	sprintf( b,"                display: flex;\n");	p = strcat( p, b);
	sprintf( b,"                flex-wrap: nowrap;\n");	p = strcat( p, b);
	sprintf( b,"                justify-content: space-between\n");	p = strcat( p, b);
	sprintf( b,"            }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"            button, .button {\n");	p = strcat( p, b);
	sprintf( b,"                display: block;\n");	p = strcat( p, b);
	sprintf( b,"                margin: 5px;\n");	p = strcat( p, b);
	sprintf( b,"                padding: 0 12px;\n");	p = strcat( p, b);
	sprintf( b,"                border: 0;\n");	p = strcat( p, b);
	sprintf( b,"                line-height: 28px;\n");	p = strcat( p, b);
	sprintf( b,"                cursor: pointer;\n");	p = strcat( p, b);
	sprintf( b,"                color: #fff;\n");	p = strcat( p, b);
	sprintf( b,"                background: #ff3034;\n");	p = strcat( p, b);
	sprintf( b,"                border-radius: 5px;\n");	p = strcat( p, b);
	sprintf( b,"                font-size: 16px;\n");	p = strcat( p, b);
	sprintf( b,"                outline: 0\n");	p = strcat( p, b);
	sprintf( b,"            }\n");	p = strcat( p, b);
	sprintf( b,"            text, .text {\n");	p = strcat( p, b);
	sprintf( b,"                display: block;\n");	p = strcat( p, b);
	sprintf( b,"                margin: 5px;\n");	p = strcat( p, b);
	sprintf( b,"                padding: 0 12px;\n");	p = strcat( p, b);
	sprintf( b,"                border: 0;\n");	p = strcat( p, b);
	sprintf( b,"                line-height: 28px;\n");	p = strcat( p, b);
	sprintf( b,"                cursor: pointer;\n");	p = strcat( p, b);
	sprintf( b,"                color: #fff;\n");	p = strcat( p, b);
	sprintf( b,"                border-radius: 5px;\n");	p = strcat( p, b);
	sprintf( b,"                font-size: 16px;\n");	p = strcat( p, b);
	sprintf( b,"                outline: 0\n");	p = strcat( p, b);
	sprintf( b,"            }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"            .save {\n");	p = strcat( p, b);
	sprintf( b,"                position: absolute;\n");	p = strcat( p, b);
	sprintf( b,"                right: 120px;\n");	p = strcat( p, b);
	sprintf( b,"                top: 5px;\n");	p = strcat( p, b);
	sprintf( b,"                height: 16px;\n");	p = strcat( p, b);
	sprintf( b,"                line-height: 16px;\n");	p = strcat( p, b);
	sprintf( b,"                padding: 0 4px;\n");	p = strcat( p, b);
	sprintf( b,"                text-decoration: none;\n");	p = strcat( p, b);
	sprintf( b,"                cursor: pointer\n");	p = strcat( p, b);
	sprintf( b,"            }\n");	p = strcat( p, b);
	sprintf( b,"            .startstop {\n");	p = strcat( p, b);
	sprintf( b,"                position: absolute;\n");	p = strcat( p, b);
	sprintf( b,"                right: 5px;\n");	p = strcat( p, b);
	sprintf( b,"                top: 5px;\n");	p = strcat( p, b);
	sprintf( b,"                height: 16px;\n");	p = strcat( p, b);
	sprintf( b,"                line-height: 16px;\n");	p = strcat( p, b);
	sprintf( b,"                padding: 0 4px;\n");	p = strcat( p, b);
	sprintf( b,"                text-decoration: none;\n");	p = strcat( p, b);
	sprintf( b,"                cursor: pointer\n");	p = strcat( p, b);
	sprintf( b,"            }\n");	p = strcat( p, b);
	sprintf( b,"            .status {\n");	p = strcat( p, b);
	sprintf( b,"                position: absolute;\n");	p = strcat( p, b);
	sprintf( b,"                right: 240px;\n");	p = strcat( p, b);
	sprintf( b,"                top: 5px;\n");	p = strcat( p, b);
	sprintf( b,"                height: 16px;\n");	p = strcat( p, b);
	sprintf( b,"                line-height: 16px;\n");	p = strcat( p, b);
	sprintf( b,"                padding: 0 4px;\n");	p = strcat( p, b);
	sprintf( b,"                text-decoration: none;\n");	p = strcat( p, b);
	sprintf( b,"                cursor: pointer\n");	p = strcat( p, b);
	sprintf( b,"            }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"            button:hover {\n");	p = strcat( p, b);
	sprintf( b,"                background: #ff49d8\n");	p = strcat( p, b);
	sprintf( b,"            }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"            button:active {\n");	p = strcat( p, b);
	sprintf( b,"                background: #f21c21\n");	p = strcat( p, b);
	sprintf( b,"            }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"            button.disabled {\n");	p = strcat( p, b);
	sprintf( b,"                cursor: default;\n");	p = strcat( p, b);
	sprintf( b,"                background: #a0a0a0\n");	p = strcat( p, b);
	sprintf( b,"            }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"            .image-container {\n");	p = strcat( p, b);
	sprintf( b,"                position: relative;\n");	p = strcat( p, b);
	sprintf( b,"                min-width: 160px\n");	p = strcat( p, b);
	sprintf( b,"            }\n");	p = strcat( p, b);
	sprintf( b,"        </style>\n");	p = strcat( p, b);
	sprintf( b,"    </head>\n");	p = strcat( p, b);
	sprintf( b,"    <body>\n");	p = strcat( p, b);
	sprintf( b,"        <section class=\"main\">\n");	p = strcat( p, b);
	sprintf( b,"            <div id=\"content\">\n");	p = strcat( p, b);
	sprintf( b,"                <figure>\n");	p = strcat( p, b);
	sprintf( b,"                    <div id=\"stream-container\" class=\"image-container\">\n");	p = strcat( p, b);
	sprintf( b,"                        <div id=\"data-status\" class=\"text status\">123456789</div>\n");	p = strcat( p, b);
	sprintf( b,"                        <a id=\"save-still\" href=\"#\" class=\"button save\" download=\"capture.jpg\">Save</a>\n");	p = strcat( p, b);
	sprintf( b,"                        <div id=\"toggle-stream\" class=\"button startstop\">Start Stream</div>\n");	p = strcat( p, b);
	sprintf( b,"                        <img id=\"stream\" src=\"\" crossorigin>\n");	p = strcat( p, b);
	sprintf( b,"                    </div>\n");	p = strcat( p, b);
	sprintf( b,"                </figure>\n");	p = strcat( p, b);
	sprintf( b,"            </div>\n");	p = strcat( p, b);
	sprintf( b,"        </section>\n");	p = strcat( p, b);
	sprintf( b,"        <script src=\"index.js\"></script>\n");	p = strcat( p, b);
	sprintf( b,"    </body>\n");	p = strcat( p, b);
	sprintf( b,"</html>\n");	p = strcat( p, b);

	return get_buf_size();
}

int	get_java( void)
{
	char	b[ 1024];
	char	*p;

	p = index_buf;
	p[ 0] = 0;

	sprintf( b,"document.addEventListener('DOMContentLoaded', function (event) {\n");	p = strcat( p, b);
	sprintf( b,"  var baseHost = document.location.origin\n");	p = strcat( p, b);
	sprintf( b,"  var streamUrl = baseHost + ':81'\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"  function fetchUrl(url, cb){\n");	p = strcat( p, b);
	sprintf( b,"    fetch(url)\n");	p = strcat( p, b);
	sprintf( b,"      .then(function (response) {\n");	p = strcat( p, b);
	sprintf( b,"        if (response.status !== 200) {\n");	p = strcat( p, b);
	sprintf( b,"          cb(response.status, response.statusText);\n");	p = strcat( p, b);
	sprintf( b,"        } else {\n");	p = strcat( p, b);
	sprintf( b,"          response.text().then(function(data){\n");	p = strcat( p, b);
	sprintf( b,"            cb(200, data);\n");	p = strcat( p, b);
	sprintf( b,"          }).catch(function(err) {\n");	p = strcat( p, b);
	sprintf( b,"            cb(-1, err);\n");	p = strcat( p, b);
	sprintf( b,"          });\n");	p = strcat( p, b);
	sprintf( b,"        }\n");	p = strcat( p, b);
	sprintf( b,"      })\n");	p = strcat( p, b);
	sprintf( b,"      .catch(function(err) {\n");	p = strcat( p, b);
	sprintf( b,"        cb(-1, err);\n");	p = strcat( p, b);
	sprintf( b,"      });\n");	p = strcat( p, b);
	sprintf( b,"  }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"  const hide = el => {\n");	p = strcat( p, b);
	sprintf( b,"    el.classList.add('hidden')\n");	p = strcat( p, b);
	sprintf( b,"  }\n");	p = strcat( p, b);
	sprintf( b,"  const show = el => {\n");	p = strcat( p, b);
	sprintf( b,"    el.classList.remove('hidden')\n");	p = strcat( p, b);
	sprintf( b,"  }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"  const disable = el => {\n");	p = strcat( p, b);
	sprintf( b,"    el.classList.add('disabled')\n");	p = strcat( p, b);
	sprintf( b,"    el.disabled = true\n");	p = strcat( p, b);
	sprintf( b,"  }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"  const enable = el => {\n");	p = strcat( p, b);
	sprintf( b,"    el.classList.remove('disabled')\n");	p = strcat( p, b);
	sprintf( b,"    el.disabled = false\n");	p = strcat( p, b);
	sprintf( b,"  }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"  const updateValue = (el, value, updateRemote) => {\n");	p = strcat( p, b);
	sprintf( b,"    updateRemote = updateRemote == null ? true : updateRemote\n");	p = strcat( p, b);
	sprintf( b,"    let initialValue\n");	p = strcat( p, b);
	sprintf( b,"    if (el.type === 'checkbox') {\n");	p = strcat( p, b);
	sprintf( b,"      initialValue = el.checked\n");	p = strcat( p, b);
	sprintf( b,"      value = !!value\n");	p = strcat( p, b);
	sprintf( b,"      el.checked = value\n");	p = strcat( p, b);
	sprintf( b,"    } else {\n");	p = strcat( p, b);
	sprintf( b,"      initialValue = el.value\n");	p = strcat( p, b);
	sprintf( b,"      el.value = value\n");	p = strcat( p, b);
	sprintf( b,"    }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"    if (updateRemote && initialValue !== value) {\n");	p = strcat( p, b);
	sprintf( b,"      updateConfig(el);\n");	p = strcat( p, b);
	sprintf( b,"    } else if(!updateRemote){\n");	p = strcat( p, b);
	sprintf( b,"    }\n");	p = strcat( p, b);
	sprintf( b,"  }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"  function updateConfig (el) {\n");	p = strcat( p, b);
	sprintf( b,"    let value\n");	p = strcat( p, b);
	sprintf( b,"    switch (el.type) {\n");	p = strcat( p, b);
	sprintf( b,"      case 'checkbox':\n");	p = strcat( p, b);
	sprintf( b,"        value = el.checked ? 1 : 0\n");	p = strcat( p, b);
	sprintf( b,"        break\n");	p = strcat( p, b);
	sprintf( b,"      case 'range':\n");	p = strcat( p, b);
	sprintf( b,"      case 'select-one':\n");	p = strcat( p, b);
	sprintf( b,"        value = el.value\n");	p = strcat( p, b);
	sprintf( b,"        break\n");	p = strcat( p, b);
	sprintf( b,"      case 'button':\n");	p = strcat( p, b);
	sprintf( b,"      case 'submit':\n");	p = strcat( p, b);
	sprintf( b,"        value = '1'\n");	p = strcat( p, b);
	sprintf( b,"        break\n");	p = strcat( p, b);
	sprintf( b,"      default:\n");	p = strcat( p, b);
	sprintf( b,"        return\n");	p = strcat( p, b);
	sprintf( b,"    }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"    const query = `${baseHost}/control?var=${el.id}&val=${value}`\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"    fetch(query)\n");	p = strcat( p, b);
	sprintf( b,"      .then(response => {\n");	p = strcat( p, b);
	sprintf( b,"        console.log(`request to ${query} finished, status: ${response.status}`)\n");	p = strcat( p, b);
	sprintf( b,"      })\n");	p = strcat( p, b);
	sprintf( b,"  }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"  // read initial values\n");	p = strcat( p, b);
	sprintf( b,"  fetch(`${baseHost}/status`)\n");	p = strcat( p, b);
	sprintf( b,"    .then(function (response) {\n");	p = strcat( p, b);
	sprintf( b,"      return response.json()\n");	p = strcat( p, b);
	sprintf( b,"    })\n");	p = strcat( p, b);
	sprintf( b,"    .then(function (state) {\n");	p = strcat( p, b);
	sprintf( b,"      document\n");	p = strcat( p, b);
	sprintf( b,"        .querySelectorAll('.default-action')\n");	p = strcat( p, b);
	sprintf( b,"        .forEach(el => {\n");	p = strcat( p, b);
	sprintf( b,"          updateValue(el, state[el.id], false)\n");	p = strcat( p, b);
	sprintf( b,"        })\n");	p = strcat( p, b);
	sprintf( b,"    })\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"  const view = document.getElementById('stream')\n");	p = strcat( p, b);
	sprintf( b,"  const viewContainer = document.getElementById('stream-container')\n");	p = strcat( p, b);
	sprintf( b,"  const streamButton = document.getElementById('toggle-stream')\n");	p = strcat( p, b);
	sprintf( b,"  const saveButton = document.getElementById('save-still')\n");	p = strcat( p, b);
	sprintf( b,"  const data_status = document.getElementById('data-status')\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"  function dispData( data){\n");	p = strcat( p, b);
	sprintf( b,"    data_status.innerText = \"T:\" +data.t +\"C  \"+ \" H:\"+ data.h + \"%%RH\";\n");	p = strcat( p, b);
	sprintf( b,"    setTimeout(() => {\n");	p = strcat( p, b);
	sprintf( b,"      getData();\n");	p = strcat( p, b);
	sprintf( b,"    }, 500);\n");	p = strcat( p, b);
	sprintf( b,"}\n");	p = strcat( p, b);
	sprintf( b,"  \n");	p = strcat( p, b);
	sprintf( b," function getData( ){\n");	p = strcat( p, b);
	sprintf( b,"    fetch(`${baseHost}/data.json`)\n");	p = strcat( p, b);
	sprintf( b,"    .then( response => response.json())\n");	p = strcat( p, b);
	sprintf( b,"    .then( data => dispData( data) )\n");	p = strcat( p, b);
	sprintf( b,"    .catch((reason) => {\n");	p = strcat( p, b);
	sprintf( b,"      console.log( reason);\n");	p = strcat( p, b);
	sprintf( b,"      });\n");	p = strcat( p, b);
	sprintf( b,"    }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"  const stopStream = () => {\n");	p = strcat( p, b);
	sprintf( b,"    window.stop();\n");	p = strcat( p, b);
	sprintf( b,"    streamButton.innerHTML = 'Start Stream'\n");	p = strcat( p, b);
	sprintf( b,"    view.src = `${baseHost}/capture?_cb=${Date.now()}`\n");	p = strcat( p, b);
	sprintf( b,"    show(viewContainer)\n");	p = strcat( p, b);
	sprintf( b,"  }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"  const startStream = () => {\n");	p = strcat( p, b);
	sprintf( b,"    view.src = `${streamUrl}/stream`\n");	p = strcat( p, b);
	sprintf( b,"    show(viewContainer)\n");	p = strcat( p, b);
	sprintf( b,"    streamButton.innerHTML = 'Stop Stream'\n");	p = strcat( p, b);
	sprintf( b,"  }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"  // Attach actions to buttons\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"  streamButton.onclick = () => {\n");	p = strcat( p, b);
	sprintf( b,"    const streamEnabled = streamButton.innerHTML === 'Stop Stream'\n");	p = strcat( p, b);
	sprintf( b,"    if (streamEnabled) {\n");	p = strcat( p, b);
	sprintf( b,"      stopStream()\n");	p = strcat( p, b);
	sprintf( b,"    } else {\n");	p = strcat( p, b);
	sprintf( b,"      startStream()\n");	p = strcat( p, b);
	sprintf( b,"    }\n");	p = strcat( p, b);
	sprintf( b,"  }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"  saveButton.onclick = () => {\n");	p = strcat( p, b);
	sprintf( b,"    var canvas = document.createElement(\"canvas\");\n");	p = strcat( p, b);
	sprintf( b,"    canvas.width = view.width;\n");	p = strcat( p, b);
	sprintf( b,"    canvas.height = view.height;\n");	p = strcat( p, b);
	sprintf( b,"    document.body.appendChild(canvas);\n");	p = strcat( p, b);
	sprintf( b,"    var context = canvas.getContext('2d');\n");	p = strcat( p, b);
	sprintf( b,"    context.drawImage(view,0,0);\n");	p = strcat( p, b);
	sprintf( b,"    try {\n");	p = strcat( p, b);
	sprintf( b,"      var dataURL = canvas.toDataURL('image/jpeg');\n");	p = strcat( p, b);
	sprintf( b,"      saveButton.href = dataURL;\n");	p = strcat( p, b);
	sprintf( b,"      var d = new Date();\n");	p = strcat( p, b);
	sprintf( b,"      saveButton.download = d.getFullYear() + (\"0\"+(d.getMonth()+1)).slice(-2) + (\"0\" + d.getDate()).slice(-2) + (\"0\" + d.getHours()).slice(-2) + (\"0\" + d.getMinutes()).slice(-2) + (\"0\" + d.getSeconds()).slice(-2) + \".jpg\";\n");	p = strcat( p, b);
	sprintf( b,"    } catch (e) {\n");	p = strcat( p, b);
	sprintf( b,"      console.error(e);\n");	p = strcat( p, b);
	sprintf( b,"    }\n");	p = strcat( p, b);
	sprintf( b,"    canvas.parentNode.removeChild(canvas);\n");	p = strcat( p, b);
	sprintf( b,"  }\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"  // Attach default on change action\n");	p = strcat( p, b);
	sprintf( b,"  document\n");	p = strcat( p, b);
	sprintf( b,"    .querySelectorAll('.default-action')\n");	p = strcat( p, b);
	sprintf( b,"    .forEach(el => {\n");	p = strcat( p, b);
	sprintf( b,"      el.onchange = () => updateConfig(el)\n");	p = strcat( p, b);
	sprintf( b,"    })\n");	p = strcat( p, b);
	sprintf( b,"\n");	p = strcat( p, b);
	sprintf( b,"  getData();\n");	p = strcat( p, b);
	sprintf( b,"  startStream();\n");	p = strcat( p, b);
	sprintf( b,"})\n");	p = strcat( p, b);

	return get_buf_size();
}

int	get_data( void)
{
	char	b[ 1024];
	char	*p;

	p = index_buf;
	p[ 0] = 0;

	return get_buf_size();
}

extern	int	 index_buf_len;

static esp_err_t index_handler(httpd_req_t *req)
{
	int	len;
	
	len = get_html();
     httpd_resp_send(req, index_buf, len);
    return ESP_OK;
}

static esp_err_t java_handler(httpd_req_t *req)
{
	int	len;
	
	len = get_java();
     httpd_resp_send(req, index_buf, len);
    return ESP_OK;
}

static esp_err_t data_handler(httpd_req_t *req)
{
    static char json_response[1024];

    char *p = json_response;
    *p++ = '{';

    p += sprintf(p, "\"t\":%4.1f", temperature);
    p += sprintf(p, ",\"h\":%4.1f", humidity*100.0);
    *p++ = '}';
    *p++ = 0;
    httpd_resp_set_type(req, "application/json");
    httpd_resp_set_hdr(req, "Access-Control-Allow-Origin", "*");
    return httpd_resp_send(req, json_response, strlen(json_response));
}


void startCameraServer()
{

    httpd_config_t config = HTTPD_DEFAULT_CONFIG();
    config.max_uri_handlers = 16;

    httpd_uri_t index_uri = {
        .uri = "/",
        .method = HTTP_GET,
        .handler = index_handler,
        .user_ctx = NULL};
    httpd_uri_t java_uri = {
        .uri = "/index.js",
        .method = HTTP_GET,
        .handler = java_handler,
        .user_ctx = NULL};

    httpd_uri_t data_uri = {
        .uri = "/data.json",
        .method = HTTP_GET,
        .handler = data_handler,
        .user_ctx = NULL};

    httpd_uri_t status_uri = {
        .uri = "/status",
        .method = HTTP_GET,
        .handler = status_handler,
        .user_ctx = NULL};

    httpd_uri_t capture_uri = {
        .uri = "/capture",
        .method = HTTP_GET,
        .handler = capture_handler,
        .user_ctx = NULL};

    httpd_uri_t stream_uri = {
        .uri = "/stream",
        .method = HTTP_GET,
        .handler = stream_handler,
        .user_ctx = NULL};

    httpd_uri_t bmp_uri = {
        .uri = "/bmp",
        .method = HTTP_GET,
        .handler = bmp_handler,
        .user_ctx = NULL};
    ra_filter_init(&ra_filter, 20);

    ESP_LOGI(TAG, "Starting web server on port: '%d'", config.server_port);
    if (httpd_start(&camera_httpd, &config) == ESP_OK)
    {
        httpd_register_uri_handler(camera_httpd, &index_uri);
        httpd_register_uri_handler(camera_httpd, &java_uri);
		httpd_register_uri_handler(camera_httpd, &data_uri);
        httpd_register_uri_handler(camera_httpd, &status_uri);
        httpd_register_uri_handler(camera_httpd, &capture_uri);
        httpd_register_uri_handler(camera_httpd, &bmp_uri);
	}

    config.server_port += 1;
    config.ctrl_port += 1;
    ESP_LOGI(TAG, "Starting stream server on port: '%d'", config.server_port);
    if (httpd_start(&stream_httpd, &config) == ESP_OK)
    {
        httpd_register_uri_handler(stream_httpd, &stream_uri);
    }
}
